﻿using System;
using System.Drawing;
using System.Reflection;
using System.Text;
using MinorShift.Emuera;
using NTERA.Interop;

namespace NTERA.Game.Display
{
	//難読化用属性。enum.ToString()やenum.Parse()を行うなら(Exclude=true)にすること。
	[Obfuscation(Exclude=false)]
	internal enum DisplayLineLastState
	{
		None = 0,
		Normal = 1,
		Selected = 2,
		BackLog = 3
	}

    /// <summary>
    /// Display line. It consists of one or more buttons (ConsoleButtonString)
    /// </summary>
    public sealed class ConsoleDisplayLine
	{
		
		//public ConsoleDisplayLine(EmueraConsole parentWindow, ConsoleButtonString[] buttons, bool isLogical, bool temporary)
		public ConsoleDisplayLine(ConsoleButtonString[] buttons, bool isLogical, bool temporary)
		{
			//parent = parentWindow;
			if (buttons == null)
			{
				buttons = new ConsoleButtonString[0];
				return;
			}
			this.buttons = buttons;
			foreach (ConsoleButtonString button in buttons)
            {
                if (button == null)
                    throw new Exception();
                button.ParentLine = this;
            }
			IsLogicalLine = isLogical;
			IsTemporary = temporary;
		}
		public int LineNo = -1;
		
		///論理行の最初となる場合だけtrue。表示の都合で改行された2行目以降はfalse
		public readonly bool IsLogicalLine = true;
		public readonly bool IsTemporary;
		//EmueraConsole parent;
		ConsoleButtonString[] buttons;
		DisplayLineAlignment align;
		public ConsoleButtonString[] Buttons => buttons;
		public DisplayLineAlignment Align => align;
		bool aligned;
		public void SetAlignment(DisplayLineAlignment align)
		{
			if (aligned)
				return;
			aligned = true;
			this.align = align;
			if (buttons.Length == 0)
				return;
			//DisplayLineの幅
			int width = 0;
			foreach (ConsoleButtonString button in buttons)
				width += button.Width;
			//現在位置
			int pointX = buttons[0].PointX;

			//目標位置
			int movetoX = 0;
			if (align == DisplayLineAlignment.LEFT)
			{
				//位置固定に対応
				if (IsLogicalLine)
					return;
				movetoX = 0;
			}
			else if (align == DisplayLineAlignment.CENTER)
				movetoX = Config.WindowX / 2 - width / 2;
			else if (align == DisplayLineAlignment.RIGHT)
				movetoX = Config.WindowX - width;

			//移動距離
			int shiftX = movetoX - pointX;
			if(shiftX != 0)
				ShiftPositionX(shiftX);
		}

		public void ShiftPositionX(int shiftX)
		{
			foreach (ConsoleButtonString button in buttons)
				button.ShiftPositionX(shiftX);
		}

		public void ChangeStr(ConsoleButtonString[] newButtons)
        {
            buttons = null;
			foreach (ConsoleButtonString button in newButtons)
				button.ParentLine = this;
			buttons = newButtons;
        }

		public void Clear(Brush brush, Graphics graph, int pointY)
		{
            Rectangle rect = new Rectangle(0, pointY, Config.WindowX, Config.LineHeight);
			graph.FillRectangle(brush, rect);
		}

		//public ConsoleButtonString GetPointingButton(int pointX)
		//{
		//	////1815 優先順位を逆順にする
		//	////後から描画されるボタンが優先されるように
		//	for (int i = 0; i < buttons.Length; i++)
		//	{
		//		ConsoleButtonString button = buttons[buttons.Length - i - 1];
		//		if ((button.PointX <= pointX) && (button.PointX + button.Width >= pointX))
		//			return button;
		//	}
		//	//foreach (ConsoleButtonString button in buttons)
		//	//{
		//	//    if ((button.PointX <= pointX) && (button.PointX + button.Width >= pointX))
		//	//        return button;
		//	//}
		//	return null;
		//}

		public void DrawTo(Graphics graph, int pointY, bool isBackLog, bool force, TextDrawingMode mode)
		{
            foreach (ConsoleButtonString button in buttons)
                button.DrawTo(graph, pointY, isBackLog, mode);
		}
		
		public override string ToString()
		{
			if (buttons == null)
				return "";
			StringBuilder builder = new StringBuilder();
			foreach (ConsoleButtonString button in buttons)
				builder.Append(button);
			return builder.ToString();
		}
	}
}
